/*
  Copyright 1992 Mark Spychalla

  Permission to use, copy, modify, distribute "xdart" and
  its documentation for any purpose is hereby granted without fee,
  provided that the above copyright notice appear in all copies and that
  both that copyright notice and this permission notice appear in
  supporting documentation, and that the name of Mark Spychalla not be used
  in advertising or publicity pertaining to distribution of the software
  without specific, written prior permission.  Mark Spychalla makes no
  representations about the suitability of this software for any purpose.
  It is provided "as is" without express or implied warranty.

  Mark Spychalla disclaims all warranties with regard to this software,
  including all implied warranties of merchantability and fitness, in no
  event shall Mark Spychalla be liable for any special, indirect or
  consequential damages or any damages whatsoever resulting from loss of use,
  data or profits, whether in an action of contract, negligence or other
  tortious action, arising out of or in connection with the use or performance
  of this software.
*/

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/X.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <ctype.h>
#include <errno.h>
#include <math.h>

#define MAX_REQUEST 16384 
#define END_OF_LINE  -1 
#define END_OF_TRACE -2 
#define VOID_VALUE   -3 
#define MAXSERVERS   20 
#define MAXLINE      512
#define MAXOUTSTANDING 2 
#define HASHTABLESIZE 60017
#define NUMFDS 64
#define MAXDIST 99999
#define NEW_BUFSIZ    32200 
#define PIXELSPERLINE 490 
#define NUMSTIPPLES 17
#define STIPPLESIZE 4
#define MAXCOLORDIST (443.40501)
#define BITSPERBYTE 8
#define EOFREPLACEMENT '\007'

Window win;
Display *dpy;
GC gc, temp_gc;
Colormap colormap;
XColor xcolor;
long color;
int screen;
XEvent event;
Pixmap stipple[NUMSTIPPLES];
XVisualInfo vInfo;

static int stipples[NUMSTIPPLES][NUMSTIPPLES * 2 + 1] = { 
{0},
{1, 1, 1},
{2, 0, 2, 2, 1},
{3, 1, 0, 1, 2, 3, 1},
{4, 0, 1, 0, 3, 2, 1, 2, 3}, 
{5, 0, 0, 0, 2, 2, 0, 2, 3, 3, 2},
{6, 0, 2, 1, 1, 1, 3, 2, 0, 3, 1, 3, 3},
{7, 0, 1, 0, 3, 1, 0, 1, 1, 2, 2, 3, 1, 3, 3},
{8, 0, 1, 0, 3, 1, 0, 1, 2, 2, 1, 2, 3, 3, 0, 3, 2},
{9, 0, 0, 0, 2, 1, 2, 1, 3, 2, 0, 2, 1, 2, 3, 3, 1, 3, 2},
{10,0, 0, 0, 1, 0, 3, 1, 0, 1, 2, 2, 1, 2, 2, 2, 3, 3, 0, 3, 2},
{11,0, 1, 0, 3, 1, 0, 1, 1, 1, 2, 1, 3, 2, 1, 2, 2, 3, 0, 3, 2, 3, 3},
{12,0, 0, 0, 2, 1, 0, 1, 1, 1, 2, 1, 3, 2, 1, 2, 3, 3, 0, 3, 1, 3, 2, 3, 3},
{13,0, 0, 0, 2, 0, 3, 1, 0, 1, 1, 1, 2, 2, 0, 2, 2, 2, 3, 3, 0, 3, 1, 3, 2,
 3, 3},
{14,0, 0, 0, 1, 0, 2, 0, 3, 1, 0, 1, 1, 1, 3, 2, 1, 2, 2, 2, 3, 3, 0, 3, 1, 
 3, 2, 3, 3},
{15,0, 0, 0, 1, 0, 2, 0, 3, 1, 0, 1, 2, 1, 3, 2, 0, 2, 1, 2, 2, 2, 3, 3, 0,
 3, 1, 3, 2, 3, 3},
{16,0, 0, 0, 1, 0, 2, 0, 3, 1, 0, 1, 1, 1, 2, 1, 3, 2, 0, 2, 1, 2, 2, 2, 3,
 3, 0, 3, 1, 3, 2, 3, 3}
};

char bits[(STIPPLESIZE * STIPPLESIZE) / BITSPERBYTE];

static struct timeval timeout = { 0, 500000 }; /* one half second */
fd_set mask, readMask;
char *readbuf[MAXSERVERS], *writebuf[MAXSERVERS];
FILE *readfd[MAXSERVERS], *writefd[MAXSERVERS], *errorfd;
char line[MAXLINE];
int sockfd[MAXSERVERS], numServers, width, numberFound, errno, x_pos, y_pos;
int Size, oldSize, Xpos, Ypos, mSize, mOldSize, screenWidth, screenHeight;
int done, quit, needInit, frameBufferLen, counter, counter2, numColors;
int numBits, initialSize, optval;

typedef struct HashTable hashTableElement;

struct HashTable{
   long color;
   hashTableElement *next;
   int value;
   short red, green, blue;
};

hashTableElement *hashtable[HASHTABLESIZE];
hashTableElement colors[256];

struct FrameBuffer{
   short red, green, blue;
};

struct FrameBuffer *frameBuffer;


void sendFile(fd)
FILE *fd;
/******************************************************************************/
/* Read the contents of the FILE *fd, write each line to all the servers.     */
/* Return to caller when an EOF is encountered on the input file.             */
/******************************************************************************/
{
int index;

/* Write each of the line to all the servers */

   while(fgets(line, MAXLINE, fd) != NULL){
      for(index = 0; index < numServers; index++){
         if(fputs(line, writefd[index]) < 0){
            fprintf(stderr, "str_cli: puts error on writeFd\n");
            exit(1);
            }
         }
      }


/* Non-buffering sockets please */

      optval = 0;

      for(index = 0; index < numServers; index++){
         if(setsockopt(sockfd[index], SOL_SOCKET, TCP_NODELAY, (char *)&optval,
         sizeof(int)) < 0){
            fprintf(stderr, "setsockopt failed.");
            exit(1);
            }
         }

/* Tell server we are done parsing */

   for(index = 0; index < numServers; index++){
      if(fputc(EOFREPLACEMENT, writefd[index]) < 0){
         fprintf(stderr, "STR_CLI: putc error on writeFd\n");
         exit(1);
         }
      fflush(writefd[index]);
      }

/* Any problems? */

   if(ferror(fd)){
      fprintf(stderr, "str_cli: error reading file");
      exit(1);
      }
}



void OpenSockets(serversFile)
char *serversFile;
/******************************************************************************/
/* Initiate connections with all our servers.                                 */
/******************************************************************************/
{
struct sockaddr_in serv_addr;
FILE *fd;
struct hostent *hostp;
int index, port;

/* Servers file exist? */

   if((fd = fopen(serversFile, "r")) == NULL){
      fprintf(stderr, "File %s not found\n", serversFile);
      exit(1);
      }

   if(fscanf(fd, "%d", &port) != 1){
      fprintf(stderr, "No port specified in %s\n", serversFile);
      exit(1);
      }
   
   numServers = 0;

   while(fgets(line, MAXLINE, fd) != NULL){

/* Figure out server's address */

      if(strlen(line) > 0){
         line[strlen(line) - 1] = '\0';
         }

      if((hostp = gethostbyname(line)) != 0) {

/* Fill in the structure "serv_addr" with the address of the server
   we want to connect with.  */

         bzero((char *)&serv_addr, sizeof(serv_addr));
         serv_addr.sin_family      = AF_INET;
         bcopy(hostp->h_addr, (char *) &serv_addr.sin_addr, hostp->h_length);
         serv_addr.sin_port        = htons(port);

/* Open a TCP socket (an Internet stream socket). */

         if((sockfd[numServers] = socket(AF_INET, SOCK_STREAM, 0)) < 0){
            fprintf(stderr, "client: can't open stream socket");
            exit(1);
            }

/* Connect to the server */

         if(connect(sockfd[numServers], (struct sockaddr *)&serv_addr,
         sizeof(serv_addr)) < 0){
            fprintf(stderr, "client: can't connect to server %s\n", line);
            exit(1);
            }

         numServers++;
         }
      }
   fclose(fd);

/* Initialize various values */
 
   printf("numServers = %d\n", numServers);
   printf("Enter q to quit\n\n");

   errorfd = NULL;
   done = 0;
   quit = 0;
   needInit = 1;
   x_pos = 0;
   y_pos = 0;
   Xpos = 0;
   Ypos = 0;
   numColors = 0;
   frameBuffer = NULL;
   counter2 = 0;
   frameBufferLen = 3;
   memset(hashtable, NULL, HASHTABLESIZE * sizeof(hashTableElement *)); 

/* Set the bits in the mask corresponding to the sockets and file */
/* descriptors we are interested in                               */

   FD_ZERO(&mask);
   FD_SET(fileno(stdin), &mask);

   for(index = 0; index < numServers; index++){ 
      readfd[index] = fdopen(sockfd[index], "r");
      if((readbuf[index] = (char *)calloc(1, NEW_BUFSIZ)) == NULL){
         fprintf(stderr, "Failed calloc()\n");
         exit(1);
         }
      setvbuf(readfd[index], readbuf[index], _IOFBF, NEW_BUFSIZ);
      writefd[index] = fdopen(sockfd[index], "w");
      if((writebuf[index] = (char *)calloc(1, NEW_BUFSIZ)) == NULL){
         fprintf(stderr, "Failed calloc()\n");
         exit(1);
         }
      setvbuf(writefd[index], writebuf[index], _IOFBF, NEW_BUFSIZ);
      FD_SET(sockfd[index], &mask);
      }
}



void handleOutput(fd)
FILE *fd;
/******************************************************************************/
/* Get pixels back from a server.                                             */
/******************************************************************************/
{

   if(counter < frameBufferLen){
      do{

/* Assign current point to be traced */

         fprintf(fd, " %d %d", x_pos, y_pos);
         counter++;

/* Figure out next point in ordered sequence */

         do{
            x_pos += Size;
         }while(((x_pos % oldSize) == 0) && ((y_pos % oldSize) == 0));

         if(x_pos >= screenWidth){
            x_pos = 0;
            y_pos += Size;
            if(y_pos >= screenHeight){
               y_pos = 0;
               oldSize = Size;
               Size = Size / 2;

               do{
                  x_pos += Size;
               }while((x_pos % oldSize) == 0);
      
               }
      
            if((x_pos == 0) && ((y_pos % oldSize) == 0))
               x_pos+=Size;

            }

      }while(((counter % PIXELSPERLINE) != 0) &&
      (counter < (frameBufferLen - 1)));

/* Are we done? */

      if(counter >= (frameBufferLen - 1)){

/* Send out last point to be traced */

         fprintf(fd, " %d %d", x_pos, y_pos);
         counter++;

         done = 1;
         }

/* Tell server to send back point color values */

      fprintf(fd, " %d %d\n", END_OF_LINE, END_OF_LINE);
      fflush(fd);
      }
}



void initScreen()
/******************************************************************************/
/* Initialize the screen.                                                     */
/******************************************************************************/
{
int index, index2;

/* Can we connect with the X server? */

   if((dpy = XOpenDisplay(NULL)) == NULL){
      fprintf(stderr, "Cannot connect with X server.\n");
      exit(0);
      }

   screen = DefaultScreen(dpy);

/* OK, what do we get for a screen? */

/* Default monochrome */

   numBits = 1;

/* A 24 bit TrueColor ? */

   if(XMatchVisualInfo(dpy, screen, 24, TrueColor, &vInfo)){
      numBits = 24;
   }else{

/* An 8 bit PseudoColor ? */

      if(XMatchVisualInfo(dpy, screen, 8, PseudoColor, &vInfo)){
         numBits = 8;
         }
      }

/* Make a vanilla window */

   win = XCreateSimpleWindow(dpy, RootWindow(dpy,screen), 0, 0,
      screenWidth, screenHeight, 0, 0, 0);

/* Some window managers are not friendly, explicitly set the background color */

   XSetWindowBackground(dpy, win, (long)BlackPixel(dpy, DefaultScreen(dpy)));

/* Initialize our copy of the frame buffer */
 
   frameBufferLen = screenWidth * screenHeight;
   frameBuffer    = (struct FrameBuffer *)malloc(frameBufferLen *
   sizeof(struct FrameBuffer));

   for(index = 0; index < frameBufferLen; index++){
      frameBuffer[index].red = VOID_VALUE;
      }

   if(screenWidth > screenHeight){
      index = screenWidth;
   }else{
      index = screenHeight;
      }

   oldSize = 2;
   Size = 1;

   while(oldSize < index){
      Size = oldSize;
      oldSize = Size * 2;
      }

   mSize = Size;
   mOldSize = oldSize;
   initialSize = Size;

   counter = 0;
   counter2 = 0;

/* Make our graphics context */

   gc = XCreateGC(dpy, win, 0x0, NULL);

/* Some window managers are not friendly, explicitly set the background color */

   XSetWindowBackground(dpy, win, (long)BlackPixel(dpy, DefaultScreen(dpy)));
   XSetBackground(dpy, gc, (long)BlackPixel(dpy, DefaultScreen(dpy)));

/* Name our window */

   XStoreName(dpy, win, "xdart");

   if(numBits == 1){

      XSetFillStyle(dpy, gc, FillOpaqueStippled);

/* Create Tiles for monochrome display */

      for(index = 0; index < NUMSTIPPLES; index++){
         stipple[index]= XCreateBitmapFromData(dpy, win, bits, STIPPLESIZE,
         STIPPLESIZE);
         temp_gc = XCreateGC(dpy, stipple[index], 0x0, NULL);
         XSetForeground(dpy, temp_gc, 0);
         XFillRectangle(dpy, stipple[index], temp_gc, 0, 0, STIPPLESIZE,
         STIPPLESIZE);
         XSetForeground(dpy, temp_gc, 1);
         for(index2 = 0; index2 < stipples[index][0]; index2++){
            XDrawPoint(dpy, stipple[index], temp_gc,
            stipples[index][index2 * 2 + 1], stipples[index][index2 * 2 + 2]);
            }
         XFreeGC(dpy, temp_gc);
         }
      }

   if(numBits == 8){

/* Make our colormap */

      colormap = XCreateColormap(dpy, win, DefaultVisual(dpy, screen),
      AllocAll);

/* Set the colormap for this window */

      XSetWindowColormap(dpy, win, colormap);

      for(index = 0; index < 256; index++){
         xcolor.red   = (short)(index << 8);
         xcolor.green = (short)(index << 8);
         xcolor.blue  = (short)(index << 8);
         xcolor.pixel = (long)index;
         xcolor.flags = DoRed | DoGreen | DoBlue;
         XStoreColor(dpy, colormap, &xcolor);
         }
      }

/* Make the window appear. */

   XMapWindow(dpy, win);
   XSync(dpy, 0);
   
/* We only want certain kinds of events */

   XSelectInput(dpy, win, ExposureMask);

/* Give each server a few extra starter lines to "fill the pipe" */
  
   for(index2 = 0; index2 < MAXOUTSTANDING; index2++){ 
      for(index = 0; index < numServers; index++){
         handleOutput(writefd[index]);
         }
      }
}



int match(r, g, b)
int r, g, b;
/******************************************************************************/
/* Pick a color from the colormap for this pixel's values.                    */
/******************************************************************************/
{
unsigned int index;
unsigned int dist1, dist2, red_diff, green_diff, blue_diff;
unsigned int color1, value, hash_value;
hashTableElement *ptr;

   switch(numBits){
      case 24:
         return ((r << 16) | (g << 8) | b);
      case  8:

/* restrict the precision of the colors we hash */

         if(r & 1)
            r--;

         if(g & 1)
            g--;

         if(b & 1)
            b--;

/* Try hashing for the color */

         value = (r * 256  + g) * 256 + b;
         hash_value = value % HASHTABLESIZE;
         ptr = hashtable[hash_value];

         while((ptr != NULL) && (ptr->value != value))
            ptr = ptr->next;

         if((ptr != NULL) && (ptr->value == value)){
               color1 = ptr->color;
         }else{

/* No entry for it then add it */

            if(numColors < 256){

               xcolor.red   = (short)(r << 8);
               xcolor.green = (short)(g << 8);
               xcolor.blue  = (short)(b << 8);
               xcolor.pixel = (long)numColors;
               xcolor.flags = DoRed | DoGreen | DoBlue;
               XStoreColor(dpy, colormap, &xcolor);

               colors[numColors].color = numColors; 
               colors[numColors].value = value; 
               colors[numColors].red   = r; 
               colors[numColors].green = g; 
               colors[numColors].blue  = b; 
               
               color1 = numColors;
               numColors++;

            }else{

/* Full color table, find closest color */

               dist1 = MAXDIST;
               color1 = 0;

               for (index = 0; index <= numColors; index++)
                  {
                  red_diff   = colors[index].red   - r;
                  green_diff = colors[index].green - g;
                  blue_diff  = colors[index].blue  - b;
                  dist2 = red_diff * red_diff +
                        green_diff * green_diff +
                         blue_diff * blue_diff;

                  if (dist2 < dist1)
                     {
                     color1 = index;
                     dist1 = dist2;
                     }
                  }
               }

               hashtable[hash_value] = (hashTableElement *)
               (calloc(1, sizeof(hashTableElement)));
               hashtable[hash_value]->next = ptr;
               hashtable[hash_value]->value = value;
               hashtable[hash_value]->color = color1;
               hashtable[hash_value]->red   = r;
               hashtable[hash_value]->green = g;
               hashtable[hash_value]->blue  = b;
            }
         return(color1);

      case 1:

         index = (int)((double)(NUMSTIPPLES)*
         ((double)sqrt((double)((double)r * (double)r +
          (double)g * (double)g + (double)b * (double)b)) / MAXCOLORDIST));

         XSetStipple(dpy, gc, stipple[index]);
         return(WhitePixel(dpy, screen));
      default:
         return(WhitePixel(dpy, screen));
         break;
      }
}



int handleInput(fd)
FILE *fd;
/******************************************************************************/
/* Get pixels back from a server.                                             */
/******************************************************************************/
{
char buffer[MAX_REQUEST];
int h, v, red, green, blue;
struct FrameBuffer *fb;

  for(;;){

/* Try to get the next pixel */

   if(fscanf(fd, " %d %d %d %d %d", &h, &v, &red, &green, &blue) <= 0){

/* Error, try again on the next line */

      if((feof(fd) != 0) || (ferror(fd) != 0)){
         if(errorfd == NULL){
            errorfd = fd;
            }
         if((fd == errorfd) && (needInit == 1)){
            quit = 1;
            done = 1;
            return -1;
         }else{
            return -2;
            }
      }else{
         if(errorfd == NULL){
            errorfd = fd;
            }
         fgets(buffer, MAX_REQUEST, fd);
         if(fd == errorfd){
            fputs(buffer, stderr);
            }
         }
   }else{
      switch(h){

         case END_OF_LINE:
            screenWidth = red;
            screenHeight = green;
            if(needInit){
               initScreen();
               needInit = 0;
               }
            fgets(buffer, MAX_REQUEST, fd);
            return 0;
            break;

         default:
            fb = &(frameBuffer[h + v * screenWidth]);
            fb->red   = red;
            fb->green = green;
            fb->blue  = blue;
            break;
         }
      }
   }
}



void UpdateImage()
/******************************************************************************/
/* Update the screen if necessary.                                            */
/******************************************************************************/
{
struct FrameBuffer *fb;
long colorValue;

   if(frameBuffer != NULL){

      fb = &(frameBuffer[Xpos + Ypos * screenWidth]);

      while((fb->red != VOID_VALUE) && (counter2 < (frameBufferLen - 1))){

         colorValue = match(fb->red, fb->green, fb->blue);
         XSetForeground(dpy, gc, (long)colorValue);
         XFillRectangle(dpy, win, gc, Xpos, Ypos, mSize, mSize);
         counter2++;

         do{
            Xpos += mSize;
         }while(((Xpos % mOldSize) == 0) && ((Ypos % mOldSize) == 0));

         if(Xpos >= screenWidth){
            Xpos = 0;
            Ypos += mSize;
            if(Ypos >= screenHeight){
               Ypos = 0;
               mOldSize = mSize;
               mSize = mSize / 2;

               do{
                  Xpos += mSize;
               }while((Xpos % mOldSize) == 0);
      
               }

            if((Xpos == 0) && ((Ypos % mOldSize) == 0))
               Xpos+=mSize;
            }

         fb = &(frameBuffer[Xpos + Ypos * screenWidth]);
         }
      }
}



void CleanUp(filename)
char *filename;
/******************************************************************************/
/* Clean up and exit.                                                         */
/******************************************************************************/
{
struct FrameBuffer *fb;
int index;
FILE *fd;

/* Close all our sockets */

   for(index = 0; index < numServers; index++){
      close(sockfd[index]);
      }

/* Write out file */

   if(needInit == 0){

      sprintf(line, "%s.rgb", filename);

      if((fd = fopen(line, "w")) == NULL){
         fprintf(stderr, "Could not open %s\n", line);
         exit(1);
         }

      for(index = 0; index < frameBufferLen; index++){
         fb = &(frameBuffer[index]);
         fprintf(fd, "%c%c%c", (unsigned char)fb->red, 
         (unsigned char)fb->green, (unsigned char)fb->blue);
         }

      fclose(fd);
      }

   printf("\nDONE.\n");
   exit(0);
}



void main(argc, argv)
int argc;
char *argv[];
/******************************************************************************/
/* Main.                                                                      */
/******************************************************************************/
{
FILE *fd;
int index;

   if(argc != 2){
      fprintf(stderr, "usage: %s <file>\n", argv[0]);
      exit(1);
      }

   if((fd = fopen(argv[1], "r")) == NULL){
      fprintf(stderr, "File %s not found\n", argv[1]);
      exit(1);
      }

/* Connect to all our servers */ 

   OpenSockets("xdart.servers");

   if(numServers < 1){
      fprintf(stderr, "No servers specified.\n");
      exit(1);
      }

/* Send out the file to be parsed by everybody */

   sendFile(fd);

/* How many file descriptor are there? */

   width = NUMFDS; 

/* until we are done with the image */

   while((counter2 < (frameBufferLen - 1)) && (quit == 0)){

/* Wait for something to happen */

      readMask = mask;

#ifdef _HPUX_SOURCE
      numberFound = select(width,(int *)&readMask, (int *)0, (int *)0,
      &timeout);
#else
      numberFound = select(width, &readMask, (fd_set *)0, (fd_set *)0,
      &timeout);
#endif


/* Need to redraw the entire screen? */

      if((needInit == 0) && (XEventsQueued(dpy, QueuedAfterFlush))){
         XNextEvent(dpy, &event);
         if(event.type == Expose){
            mSize = initialSize;
            mOldSize = mSize * 2;
            Xpos = 0;
            Ypos = 0;
            counter2 = 0;
            }
         }

/* Something bad happen? */

      if(numberFound < 0){

         if (errno == EINTR) {

/* just an interupted select() */

         }else{
            perror("select");
            exit(1);
         }

      }else{
         if(numberFound == 0) {

/* Timer expired */

         }else{
            if(FD_ISSET(fileno(stdin), &readMask)){
               fgets(line, MAXLINE, stdin);
               if(strncmp("q", line, 1) == 0){
                  done = 1;
                  quit = 1;
                  }
            }else{

/* Anything happen on the sockets? */

               for(index = 0; index < numServers; index++){
                  if(FD_ISSET(sockfd[index], &readMask)){
                     if(handleInput(readfd[index]) == 0){
                        if(done == 0){
                           handleOutput(writefd[index]);
                           }
                        }
                     }
                  }
               }
            }
         }
      UpdateImage();   
      } 

   CleanUp(argv[1]);
}
